import base64
import json
import traceback

from infoworks.sdk.url_builder import list_sources_url, create_workflow_url, list_domains_url, configure_workflow_url,list_pipeline_versions_url,list_pipelines_url,list_users_url
from infoworks.sdk.utils import IWUtils
import sys
import configparser
import requests
from infoworks.sdk.cicd.upload_configurations.domains import Domain
from infoworks.core.iw_authentication import get_bearer_token
from infoworks.sdk.cicd.upload_configurations.utils import Utils
from infoworks.sdk.cicd.upload_configurations.update_configurations import InfoworksDynamicAccessNestedDict
from infoworks.sdk.cicd.upload_configurations.local_configurations import PRE_DEFINED_MAPPINGS

class Workflow:
    def __init__(self, workflow_configuration_path, replace_words=""):
        with open(workflow_configuration_path, 'r') as file:
            json_string = file.read()
            if replace_words != "":
                for key, value in [item.split("->") for item in replace_words.split(";")]:
                    json_string = json_string.replace(key, value)
        self.configuration_obj = IWUtils.ejson_deserialize(json_string)

    def update_mappings_for_configurations(self, mappings):
        config = configparser.ConfigParser()
        config.read_dict(mappings)
        d = InfoworksDynamicAccessNestedDict(self.configuration_obj)
        for section in config.sections():
            if section in PRE_DEFINED_MAPPINGS:
                continue
            try:
                final = d.setval(section.split("$"), dict(config.items(section)))
            except KeyError as e:
                pass
        self.configuration_obj = d.data
        iw_mappings = self.configuration_obj.get("configuration", {}).get("iw_mappings", [])
        try:
            if "domain_name_mappings" in config.sections():
                domain_mappings = dict(config.items("domain_name_mappings"))
                if domain_mappings!={}:
                    for mapping in iw_mappings:
                        domain_name = mapping.get("recommendation", {}).get("domain_name", "")
                        if domain_name != "" and domain_mappings != {}:
                            mapping["recommendation"]["domain_name"] = domain_mappings.get(domain_name.lower(), domain_name)
                    self.configuration_obj["configuration"]["iw_mappings"]=iw_mappings
        except Exception as e:
            print("Failed while doing the domain mappings")
            print(str(e))
            print(traceback.format_exc())
        # handle any other generic name mappings like iw_mappings$recommendation$source_name
        try:
            generic_mappings = [i for i in config.sections() if i.lower().startswith("iw_mappings$")]
            for mapping in generic_mappings:
                lineage_list = mapping.split("$")
                lineage_list.remove("iw_mappings")
                if "recommendation" in lineage_list:
                    lineage_list.remove("recommendation")
                artifact_type = lineage_list[0]
                artifact_mappings = dict(config.items(f"iw_mappings$recommendation${artifact_type}"))
                if artifact_mappings != {}:
                    for mapping in iw_mappings:
                        artifact_name = mapping.get("recommendation", {}).get(artifact_type, "")
                        if artifact_name != "" and artifact_mappings != {}:
                            mapping["recommendation"][artifact_type] = artifact_mappings.get(artifact_name.lower(),
                                                                                             artifact_name)
                    self.configuration_obj["configuration"]["iw_mappings"] = iw_mappings
        except Exception as e:
            print("Failed while doing the generic mappings")
            print(str(e))
            print(traceback.format_exc())

    def get_existing_domain_id(self,wf_client_obj,domain_name):
        existing_domain_id=None
        if domain_name is not None:
            domains_url_base = list_domains_url(wf_client_obj.client_config)
            filter_condition = IWUtils.ejson_serialize({"name": domain_name})
            domains_url = domains_url_base + f"?filter={{filter_condition}}".format(filter_condition=filter_condition)
            response = requests.request("GET", domains_url, headers={
                'Authorization': 'Bearer ' + wf_client_obj.client_config["bearer_token"],
                'Content-Type': 'application/json'}, verify=False)
            if response.status_code == 406:
                headers = wf_client_obj.regenerate_bearer_token_if_needed(
                    {'Authorization': 'Bearer ' + wf_client_obj.client_config["bearer_token"],
                     'Content-Type': 'application/json'})
                response = requests.request("GET", domains_url, headers=headers, verify=False)
            existing_domain_id = None
            if response is not None:
                result = response.json().get("result", [])
                if len(result) > 0:
                    existing_domain_id = result[0]["id"]
                else:
                    wf_client_obj.logger.error('Can not find domain with given name {} '.format(domain_name))
                    wf_client_obj.logger.error('Unable to create workflow')
                    print(f'Can not find domain with given name {domain_name} ')
                    print('Unable to create workflow')
                    raise Exception("Unable to create workflow")
            wf_client_obj.logger.info('domainId {}'.format(existing_domain_id))
            return existing_domain_id

    def get_existing_pipeline_id(self,wf_client_obj,domain_id,pipeline_name):
        existing_pipeline_id = None
        if domain_id is not None:
            pipelines_url_base = list_pipelines_url(wf_client_obj.client_config,domain_id=domain_id)
            filter_condition = IWUtils.ejson_serialize({"name": pipeline_name})
            pipelines_url = pipelines_url_base + f"?filter={{filter_condition}}".format(filter_condition=filter_condition)
            response = requests.request("GET", pipelines_url, headers={
                'Authorization': 'Bearer ' + wf_client_obj.client_config["bearer_token"],
                'Content-Type': 'application/json'}, verify=False)
            if response.status_code == 406:
                headers = wf_client_obj.regenerate_bearer_token_if_needed(
                    {'Authorization': 'Bearer ' + wf_client_obj.client_config["bearer_token"],
                     'Content-Type': 'application/json'})
                response = requests.request("GET", pipelines_url, headers=headers, verify=False)
            if response is not None:
                result = response.json().get("result", [])
                if len(result) > 0:
                    existing_pipeline_id = result[0]["id"]
                else:
                    wf_client_obj.logger.error('Can not find pipeline with given name {} '.format(pipeline_name))
                    wf_client_obj.logger.error('Unable to create workflow')
                    print(f'Can not find pipeline with given name {pipeline_name} ')
                    print('Unable to create workflow')
                    raise Exception("Unable to create workflow")
            wf_client_obj.logger.info('pipeline_id {}'.format(existing_pipeline_id))
            return existing_pipeline_id

    def set_active_pipeline_version_id(self,wf_client_obj):
        for mapping in self.configuration_obj["configuration"]["iw_mappings"]:
            if mapping["entity_type"] == "pipeline":
                domain_name = mapping["recommendation"]["domain_name"]
                pipeline_name = mapping["recommendation"]["pipeline_name"]
                domain_id = self.get_existing_domain_id(wf_client_obj, domain_name)
                pipeline_id = self.get_existing_pipeline_id(wf_client_obj, domain_id, pipeline_name=pipeline_name)
                filter_condition = IWUtils.ejson_serialize({"is_active": True})
                pipeline_active_version_url = list_pipeline_versions_url(wf_client_obj.client_config,domain_id=domain_id,pipeline_id=pipeline_id) + f"?filter={{filter_condition}}".format(filter_condition=filter_condition)
                response = requests.request("GET", pipeline_active_version_url,
                                            headers={'Authorization': 'Bearer ' + wf_client_obj.client_config[
                                                'bearer_token'],
                                                     'Content-Type': 'application/json'}, verify=False)
                if response.status_code == 406:
                    wf_client_obj.client_config['bearer_token'] = get_bearer_token(
                        wf_client_obj.client_config["protocol"],
                        wf_client_obj.client_config["ip"],
                        wf_client_obj.client_config["port"],
                        wf_client_obj.client_config["refresh_token"])
                    headers = IWUtils.get_default_header_for_v3(wf_client_obj.client_config['bearer_token'])
                    response = requests.request("GET", pipeline_active_version_url,
                                                headers=headers, verify=False)
                if response is not None:
                    pipeline_versions_result = response.json().get("result", [])
                    if pipeline_versions_result:
                        pipeline_version_id=pipeline_versions_result[0]["version"]
                        mapping["selected_versions"][0]["version"] = pipeline_version_id
                wf_client_obj.logger.info(response.json())
                print(response.json())

    def create(self, wf_client_obj, domain_id, domain_name):
        # update the active version of pipeline
        self.set_active_pipeline_version_id(wf_client_obj)
        sources_in_wfs = []
        workflow_name = self.configuration_obj["configuration"]["entity"]["entity_name"]
        for item in self.configuration_obj["configuration"]["iw_mappings"]:
            if item["entity_type"] == "table_group" and "source_name" in item["recommendation"]:
                sources_in_wfs.append(item["recommendation"].get("source_name"))
        filter_condition = IWUtils.ejson_serialize({"name": {"$in": sources_in_wfs}})
        source_list_url = list_sources_url(wf_client_obj.client_config)
        wf_client_obj.logger.info(f"Listing source url {source_list_url}")
        src_list_url = source_list_url + f"?filter={{filter_condition}}".format(filter_condition=filter_condition)
        response = requests.request("GET", src_list_url,
                                    headers={'Authorization': 'Bearer ' + wf_client_obj.client_config['bearer_token'],
                                             'Content-Type': 'application/json'}, verify=False)
        if response.status_code == 406:
            wf_client_obj.client_config['bearer_token'] = get_bearer_token(wf_client_obj.client_config["protocol"],
                                                                           wf_client_obj.client_config["ip"],
                                                                           wf_client_obj.client_config["port"],
                                                                           wf_client_obj.client_config["refresh_token"])
            headers = IWUtils.get_default_header_for_v3(wf_client_obj.client_config['bearer_token'])
            response = requests.request("GET", src_list_url,
                                        headers=headers, verify=False)
        wf_client_obj.logger.info(response.json())
        print(response.json())
        temp_src_ids = []
        if response.status_code == 200 and len(response.json().get("result", [])) > 0:
            result = response.json().get("result", [])
            for item in result:
                temp_src_ids.append(item["id"])
        sourceids_in_wfs = list(set(temp_src_ids))
        user_email = self.configuration_obj["user_email"]
        current_user_token = wf_client_obj.client_config.get("bearer_token","")
        parts = current_user_token.split('.')
        bearer_token_string= parts[1]
        bearer_token_string += '=' * (4 - len(parts[1]) % 4)
        bearer_token_string = base64.b64decode(bearer_token_string)
        bearer_token_json = json.loads(bearer_token_string)
        bearer_token_json = json.loads(bearer_token_json.get("sub","{}"))
        user_email = bearer_token_json.get("email",user_email)
        domain_obj = Domain(None)
        new_workflow_id = ''
        final_domain_id = None
        if domain_id is None and domain_name is None:
            wf_client_obj.logger.error('Either domainId or domain Name is required to create workflow.')
            print('Either domainId or domain Name is required to create workflow.')
            sys.exit(-1)
        if domain_name is not None and domain_id is None:
            domains_url_base = list_domains_url(wf_client_obj.client_config)
            filter_condition = IWUtils.ejson_serialize({"name": domain_name})
            domains_url = domains_url_base + f"?filter={{filter_condition}}".format(filter_condition=filter_condition)
            response = requests.request("GET", domains_url, headers={
                'Authorization': 'Bearer ' + wf_client_obj.client_config["bearer_token"],
                'Content-Type': 'application/json'}, verify=False)
            if response.status_code == 406:
                headers = wf_client_obj.regenerate_bearer_token_if_needed(
                    {'Authorization': 'Bearer ' + wf_client_obj.client_config["bearer_token"],
                     'Content-Type': 'application/json'})
                response = requests.request("GET", domains_url, headers=headers, verify=False)
            existing_domain_id = None
            if response is not None:
                result = response.json().get("result", [])
                if len(result) > 0:
                    existing_domain_id = result[0]["id"]
                    final_domain_id = existing_domain_id
                else:
                    wf_client_obj.logger.error('Can not find domain with given name {} '.format(domain_name))
                    wf_client_obj.logger.error('Unable to create workflow')
                    print(f'Can not find domain with given name {domain_name} ')
                    print('Unable to create workflow')
                    raise Exception("Unable to create workflow")
                    # wf_client_obj.logger.info('Creating a domain with given name {} '.format(domain_name))
                    # domain_id_new = domain_obj.create(wf_client_obj, domain_name)
                    # print('New domain id' + domain_id_new)
                    # final_domain_id = domain_id_new
            wf_client_obj.logger.info('domainId {}'.format(existing_domain_id))
            print(f"domainId:{existing_domain_id}")
        else:
            final_domain_id = domain_id
        wf_client_obj.logger.info('Adding user {} to domain {}'.format(user_email, final_domain_id))
        print(f'Adding user {user_email} to domain {final_domain_id}')
        domain_obj.add_user_to_domain(wf_client_obj, final_domain_id, None, user_email)
        wf_client_obj.logger.info('Adding sources {} to domain {}'.format(sourceids_in_wfs, final_domain_id))
        print(f'Adding sources {sourceids_in_wfs} to domain {final_domain_id}')
        domain_obj.add_sources_to_domain(wf_client_obj, final_domain_id, sourceids_in_wfs)
        url_for_creating_workflow = create_workflow_url(wf_client_obj.client_config, final_domain_id)
        workflow_json_object = {
            "name": workflow_name
        }
        wf_client_obj.logger.info('URL to create workflow: ' + url_for_creating_workflow)
        json_string = IWUtils.ejson_serialize(workflow_json_object)
        wf_client_obj.logger.debug(json_string)
        if json_string is not None:
            try:
                response = requests.post(url_for_creating_workflow, data=json_string,
                                         headers={
                                             'Authorization': 'Bearer ' + wf_client_obj.client_config['bearer_token'],
                                             'Content-Type': 'application/json'}, verify=False)
                if response.status_code == 406:
                    wf_client_obj.client_config['bearer_token'] = get_bearer_token(
                        wf_client_obj.client_config["protocol"],
                        wf_client_obj.client_config["ip"],
                        wf_client_obj.client_config["port"],
                        wf_client_obj.client_config["refresh_token"])
                    headers = IWUtils.get_default_header_for_v3(wf_client_obj.client_config['bearer_token'])
                    response = requests.post(url_for_creating_workflow, data=json_string, headers=headers, verify=False)
                response = IWUtils.ejson_deserialize(response.content)
                wf_client_obj.logger.debug(response)
                result = response.get('result', None)
                wf_client_obj.logger.info("result is: " + str(result))
                if result is None:
                    wf_client_obj.logger.info(
                        'Cant create workflow. {} {}'.format(response.get('message'), response.get('details')))
                    print('Cant create workflow. {} {}'.format(response.get('message'), response.get('details')))
                    wf_client_obj.logger.info('Getting the existing workflow ID with given name.')
                    print('Getting the existing workflow ID with given name.')
                    workflow_base_url = create_workflow_url(wf_client_obj.client_config, final_domain_id)
                    filter_condition = IWUtils.ejson_serialize({"name": workflow_name})
                    workflow_get_url = workflow_base_url + f"?filter={{filter_condition}}".format(
                        filter_condition=filter_condition)
                    response = requests.request("GET", workflow_get_url, headers={
                        'Authorization': 'Bearer ' + wf_client_obj.client_config['bearer_token'],
                        'Content-Type': 'application/json'}, verify=False)
                    if response.status_code == 406:
                        wf_client_obj.client_config['bearer_token'] = get_bearer_token(
                            wf_client_obj.client_config["protocol"],
                            wf_client_obj.client_config["ip"],
                            wf_client_obj.client_config["port"],
                            wf_client_obj.client_config["refresh_token"])
                        headers = IWUtils.get_default_header_for_v3(wf_client_obj.client_config['bearer_token'])
                        response = requests.request("GET", workflow_get_url, headers=headers, verify=False)
                    wf_client_obj.logger.debug(response)
                    print(response)
                    existing_workflow_id = None
                    if response.status_code == 200 and len(response.json().get("result", [])) > 0:
                        existing_workflow_id = response.json().get("result", [])[0]["id"]
                        wf_client_obj.logger.info("Workflow ID found {}".format(existing_workflow_id))
                        print("Workflow ID found {}".format(existing_workflow_id))
                    if existing_workflow_id:
                        new_workflow_id = str(existing_workflow_id)
                else:
                    new_workflow_id = result.get('id')
            except Exception as ex:
                wf_client_obj.logger.error('Response from server: {}'.format(str(ex)))
                print('Response from server: {}'.format(str(ex)))

        return new_workflow_id, final_domain_id

    def configure(self, wf_client_obj, workflow_id, domain_id):
        # Added for Workflow Tags
        # utils_obj = Utils()
        # utils_obj.replace_custom_tags_names_with_mapping(self.configuration_obj["configuration"]["workflow"],
        #                                                  wf_client_obj)
        import_config_status = []
        url_for_importing_workflow = configure_workflow_url(wf_client_obj.client_config, domain_id, workflow_id)
        json_string = IWUtils.ejson_serialize({"configuration": self.configuration_obj["configuration"]})
        response = requests.put(url_for_importing_workflow, data=json_string,
                                headers={'Authorization': 'Bearer ' + wf_client_obj.client_config['bearer_token'],
                                         'Content-Type': 'application/json'}, verify=False)
        if response.status_code == 406:
            wf_client_obj.client_config['bearer_token'] = get_bearer_token(wf_client_obj.client_config["protocol"],
                                                                           wf_client_obj.client_config["ip"],
                                                                           wf_client_obj.client_config["port"],
                                                                           wf_client_obj.client_config["refresh_token"])
            headers = IWUtils.get_default_header_for_v3(wf_client_obj.client_config['bearer_token'])
            response = requests.put(url_for_importing_workflow, data=json_string, headers=headers, verify=False)
        status_code = response.status_code
        response = IWUtils.ejson_deserialize(response.content)
        error = []
        for iw_mapping in response.get("result", {}).get("configuration", {}).get("iw_mappings", []):
            import_error = iw_mapping.get("error", "")
            if import_error:
                error.append(import_error)
        workflow_import_status = "SUCCESS" if (error == [] and status_code == 200) else "FAILED"
        workflow_import_response = response if (workflow_import_status == "SUCCESS" or status_code != 200) else error
        import_config_status.append(('workflow_import_status', workflow_import_status, workflow_import_response))
        wf_client_obj.logger.info(response)
        print(response)
        if response is not None:
            wf_client_obj.logger.info(response.get("message", "") + " Done")
            print(response.get("message", "") + " Done")
        return workflow_import_status,import_config_status


import time

from infoworks.error import PipelineError
from infoworks.sdk import url_builder, local_configurations
from infoworks.sdk.base_client import BaseClient
from infoworks.sdk.generic_response import GenericResponse
from infoworks.sdk.local_configurations import Response, ErrorCode
from infoworks.sdk.utils import IWUtils


class PipelineGroupClient(BaseClient):
    def __init__(self):
        super(PipelineGroupClient, self).__init__()

    def poll_pipeline_group_job(self, pipeline_group_id=None, job_id=None,
                                poll_timeout=local_configurations.POLLING_TIMEOUT,
                                polling_frequency=local_configurations.POLLING_FREQUENCY_IN_SEC,
                                retries=local_configurations.NUM_POLLING_RETRIES):
        """
        Function to poll the pipeline group job
        :param pipeline_group_id: ID of the pipeline group
        :type pipeline_group_id: String
        :param job_id: job_id of the pipeline group job
        :type job_id: String
        :param poll_timeout: Polling timeout(default is 1200 seconds).If -1 then till the job completes polling will be done.
        :type poll_timeout: Integer
        :param polling_frequency: Frequency of the polling(default is 15seconds)
        :type polling_frequency: Integer
        :param retries: Number of retries in case of failure(default 3)
        :type retries: Integer
        :return: response dict
        """
        failed_count = 0
        response = {}
        if poll_timeout != -1:
            timeout = time.time() + poll_timeout
        else:
            # 2,592,000 is 30 days assuming it to be max time a job can run
            timeout = time.time() + 2592000
        while True:
            if time.time() > timeout:
                break
            try:
                self.logger.info(f"Failed poll job status count: {failed_count}")
                job_monitor_url = url_builder.get_job_status_url(self.client_config, job_id)
                response = IWUtils.ejson_deserialize(self.call_api("GET", job_monitor_url,
                                                                   IWUtils.get_default_header_for_v3(
                                                                       self.client_config['bearer_token'])).content)
                result = response.get('result', {})
                if len(result) != 0:
                    job_status = result["status"]
                    print(f"pipeline_group_status : {job_status}.Sleeping for {polling_frequency} seconds")
                    self.logger.info(
                        "Job poll status : " + result["status"] + "Job completion percentage: " + str(result.get(
                            "percentage", 0)))
                    if job_status.lower() in ["completed", "failed", "aborted", "canceled"]:
                        return GenericResponse.parse_result(status=Response.Status.SUCCESS, entity_id=pipeline_group_id,
                                                            job_id=job_id, response=response)
                else:
                    self.logger.error(f"Error occurred during job {job_id} status poll")
                    if failed_count >= retries - 1:
                        return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                            error_code=ErrorCode.GENERIC_ERROR,
                                                            error_desc=f"Error occurred during job {job_id} status poll",
                                                            response=response, job_id=job_id,
                                                            entity_id=pipeline_group_id)
                    failed_count = failed_count + 1
            except Exception as e:
                self.logger.exception("Error occurred during job status poll")
                self.logger.info(str(e))
                if failed_count >= retries - 1:
                    # traceback.print_stack()
                    print(response)
                    raise PipelineError(response.get("message", "Error occurred during job status poll"))
                failed_count = failed_count + 1
            time.sleep(polling_frequency)

    def list_jobs_under_pipeline_group(self, domain_id, pipeline_group_id, params=None,pagination=True):
        """
            Function to list the jobs under pipeline group
            :param domain_id: Entity identifier for domain
            :type domain_id: String
            :param pipeline_group_id: Entity identifier of the pipeline group
            :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
            :type: JSON dict
            :param pagination: Boolean value to determine whether to return entire result set or only a portion of result defined by 'limit' under params
            :type pagination: Boolean
            :return: response dict
        """
        if None in {domain_id, pipeline_group_id}:
            self.logger.error("domain_id or pipeline_group_id cannot be None")
            raise Exception("domain_id or pipeline_group_id cannot be None")
        if params is None:
            params = {"limit": 20, "offset": 0}
        url_to_list_pipeline_grp_jobs = url_builder.get_pipeline_group_jobs_base_url(self.client_config, domain_id,
                                                                                     pipeline_group_id) + IWUtils.get_query_params_string_from_dict(
            params=params)
        pipeline_groups_jobs_list = []
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_pipeline_grp_jobs,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                initial_msg = response.get("message", "")
                while len(result) > 0:
                    pipeline_groups_jobs_list.extend(result)
                    if not pagination:
                        break
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(
                        self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                            self.client_config['bearer_token'])).content)
                    result = response.get("result", None)
                    if result is None:
                        return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                            error_code=ErrorCode.GENERIC_ERROR,
                                                            error_desc="Error in listing pipeline group jobs",
                                                            response=response
                                                            )

                response["result"] = pipeline_groups_jobs_list
                response["message"] = initial_msg
            return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
        except Exception as e:
            self.logger.error("Error in listing pipeline group jobs")
            raise PipelineError("Error in listing pipeline group jobs" + str(e))

    def trigger_pipeline_group_build(self, domain_id, pipeline_group_id, poll=True):
        """
            Function to trigger the pipeline group build
            :param domain_id: Entity identifier for domain
            :type domain_id: String
            :param pipeline_group_id: Entity identifier of the pipeline group
            :param poll: Poll job until its completion
            :type poll: Boolean
            :return: response dict
        """
        if None in {domain_id, pipeline_group_id}:
            self.logger.error("domain_id or pipeline_group_id cannot be None")
            raise Exception("domain_id or pipeline_group_id cannot be None")
        url_for_pipeline_grp_build = url_builder.get_pipeline_group_jobs_base_url(self.client_config, domain_id,
                                                                                  pipeline_group_id)
        request_body = {
            "job_type": "pipeline_group_build"
        }
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("POST", url_for_pipeline_grp_build,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                              request_body).content)
            result = response.get('result', {})
        except Exception as e:
            self.logger.error(e)
            raise PipelineError(f"Failed to submit pipeline group build job for {pipeline_group_id} " + str(e))
        if len(result) != 0 and "id" in result:
            job_id = result["id"]
            if not poll:
                self.logger.info(f"Pipeline group build job has been submitted for {pipeline_group_id} with {job_id}")
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, job_id=job_id,
                                                    entity_id=pipeline_group_id, response=response)
            else:
                return self.poll_pipeline_group_job(pipeline_group_id=pipeline_group_id, job_id=job_id)
        else:
            self.logger.error(response)
            raise PipelineError(f"Failed to submit pipeline group build job for {pipeline_group_id} ")

    def get_pipeline_group_job_details(self, domain_id, pipeline_group_id, job_id):
        """
        Gets Infoworks pipeline group job details for given pipeline group id and job id
        :param pipeline_group_id: id of the pipeline group whose details are to be fetched
        :type pipeline_group_id: String
        :param domain_id: Domain id to which the pipeline belongs to
        :type domain_id: String
        :param job_id: id of the pipeline group job
        :return: response dict
        """
        try:
            if None in {pipeline_group_id, domain_id, job_id}:
                raise Exception(f"pipeline_group_id, domain_id, job_id cannot be None")
            response = IWUtils.ejson_deserialize(self.call_api("GET", url_builder.get_pipeline_group_jobs_base_url(
                self.client_config, domain_id, pipeline_group_id) + f"/{job_id}", IWUtils.get_default_header_for_v3(
                self.client_config['bearer_token'])).content)
            result = response.get('result', {})
            if result.get('id', None) is None:
                self.logger.error('Failed to find the pipeline group job details')
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.USER_ERROR,
                                                    error_desc='Failed to find the pipeline group job details',
                                                    response=response)

            job_id = str(job_id)
            self.logger.info(
                'Successfully got the pipeline group job {id} details.'.format(id=job_id))
            return GenericResponse.parse_result(status=Response.Status.SUCCESS, entity_id=job_id,
                                                response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(e))
            self.logger.exception('Error occurred while trying to get pipeline group details.')
            raise PipelineError('Error occurred while trying to get pipeline group details.')

    def list_pipeline_job_details_in_pipeline_group_job(self, domain_id, pipeline_group_id, job_id, params=None,pagination=True):
        """
            Function to list the pipeline jobs in a pipeline group job
            :param domain_id: Entity identifier for domain
            :type domain_id: String
            :param pipeline_group_id: Entity identifier of the pipeline group
            :param job_id: Entity identifier of the pipeline group job
            :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
            :type: JSON dict
            :param pagination: Boolean value to determine whether to return entire result set or only a portion of result defined by 'limit' under params
            :type pagination: Boolean
            :return: response dict
        """
        if None in {pipeline_group_id, domain_id, job_id}:
            raise Exception(f"pipeline_group_id, domain_id, job_id cannot be None")
        if params is None:
            params = {"limit": 20, "offset": 0}
        url_to_list_pipeline_grp_jobs = url_builder.get_pipeline_group_jobs_base_url(self.client_config, domain_id,
                                                                                     pipeline_group_id) + f"/{job_id}/pipeline-jobs" + IWUtils.get_query_params_string_from_dict(
            params=params)
        pipelines_in_pipeline_groups_jobs_list = []
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_pipeline_grp_jobs,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                initial_msg = response.get("message", "")
                while len(result) > 0:
                    pipelines_in_pipeline_groups_jobs_list.extend(result)
                    if not pagination:
                        break
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(
                        self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                            self.client_config['bearer_token'])).content)
                    result = response.get("result", None)
                    if result is None:
                        return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                            error_code=ErrorCode.GENERIC_ERROR,
                                                            error_desc="Error in listing pipeline jobs in pipeline group job",
                                                            response=response
                                                            )

                response["result"] = pipelines_in_pipeline_groups_jobs_list
                response["message"] = initial_msg
            return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
        except Exception as e:
            self.logger.error("Error in listing pipeline jobs in pipeline group job")
            raise PipelineError("Error in listing pipeline jobs in pipeline group job" + str(e))

    def list_cluster_job_runs_in_pipeline_group_job(self, domain_id, pipeline_group_id, job_id, params=None, pagination=True):
        """
            Function to list the cluster jobs in a pipeline group job
            :param domain_id: Entity identifier for domain
            :type domain_id: String
            :param pipeline_group_id: Entity identifier of the pipeline group
            :param job_id: Entity identifier of the pipeline group job
            :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
            :type: JSON dict
            :param pagination: Boolean value to determine whether to return entire result set or only a portion of result defined by 'limit' under params
            :type pagination: Boolean
            :return: response dict
        """
        if None in {pipeline_group_id, domain_id, job_id}:
            raise Exception(f"pipeline_group_id, domain_id, job_id cannot be None")
        if params is None:
            params = {"limit": 20, "offset": 0}
        url_to_list_pipeline_grp_job_runs = url_builder.get_pipeline_group_jobs_base_url(self.client_config, domain_id,
                                                                                         pipeline_group_id) + f"/{job_id}/runs" + IWUtils.get_query_params_string_from_dict(
            params=params)
        cluster_jobs_list = []
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_pipeline_grp_job_runs,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                initial_msg = response.get("message", "")
                while len(result) > 0:
                    cluster_jobs_list.extend(result)
                    if not pagination:
                        break
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(
                        self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                            self.client_config['bearer_token'])).content)
                    result = response.get("result", None)
                    if result is None:
                        return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                            error_code=ErrorCode.GENERIC_ERROR,
                                                            error_desc="Error in listing cluster jobs in pipeline group job",
                                                            response=response
                                                            )

                response["result"] = cluster_jobs_list
                response["message"] = initial_msg
            return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
        except Exception as e:
            self.logger.error("Error in listing cluster jobs in pipeline group job")
            raise PipelineError("Error in listing cluster jobs in pipeline group job" + str(e))

    def get_pipeline_group_job_log_as_text(self, domain_id, pipeline_group_id, job_id):
        """
            Gets Infoworks pipeline group job logs
            :param pipeline_group_id: id of the pipeline group whose logs are to be fetched
            :type pipeline_group_id: String
            :param domain_id: Domain id to which the pipeline belongs to
            :type domain_id: String
            :param job_id: id of the pipeline group job
            :return: response dict
        """
        if None in {pipeline_group_id, domain_id, job_id}:
            raise Exception(f"pipeline_group_id, domain_id, job_id cannot be None")
        try:
            if None in {pipeline_group_id, domain_id, job_id}:
                raise Exception(f"pipeline_group_id, domain_id, job_id cannot be None")
            response = self.call_api("GET", url_builder.get_pipeline_group_jobs_base_url(
                self.client_config, domain_id, pipeline_group_id) + f"/{job_id}/logs",
                                     IWUtils.get_default_header_for_v3(
                                         self.client_config['bearer_token'])).content
            if response is None:
                self.logger.error('Failed to find the pipeline group job logs')
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.USER_ERROR,
                                                    error_desc='Failed to find the pipeline group job logs',
                                                    response=response)
            else:
                self.logger.info(
                    'Successfully got the pipeline group job {id} logs.'.format(id=job_id))
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, entity_id=job_id,
                                                    response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(e))
            self.logger.exception('Error occurred while trying to get pipeline group job logs.')
            raise PipelineError('Error occurred while trying to get pipeline group job logs.')

    def create_pipeline_group(self, pipeline_group_config):
        """
            Function to create the pipeline group
            :param pipeline_group_config: a JSON object containing pipeline group configurations
            :type pipeline_group_config: JSON Object

        pipeline_group_config_example = {
                    "name": "test_pg1",
                    "description": "",
                    "environment_id": "a801283f2e8077120d000b49",
                    "domain_id": "576a8508b25540b4b180927f",
                    "batch_engine": "SNOWFLAKE",
                    "run_job_on_data_plane": false,
                    "pipelines": [
                        {
                            "pipeline_id": "270a56ca0ae71c79039d79e0",
                            "version": 3,
                            "execution_order": 1,
                            "run_active_version": true
                        },
                        {
                            "pipeline_id": "266c068325500cdc8ccd6d69",
                            "version": 1,
                            "execution_order": 2,
                            "run_active_version": true
                        }
                    ],
                    "custom_tags": [],
                    "query_tag": "",
                    "snowflake_warehouse": "DEMO_WH"
                }
            :return: response dict
        """

        try:
            if pipeline_group_config is None or "domain_id" not in pipeline_group_config:
                raise Exception(
                    f"pipeline_config cannot be None or domain_id should be present in pipeline_group_config")
            domain_id = pipeline_group_config["domain_id"]
            response = IWUtils.ejson_deserialize(self.call_api("POST", url_builder.get_pipeline_group_base_url(
                self.client_config, domain_id), IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                                                               pipeline_group_config).content)
            result = response.get('result', {})
            pipeline_grp_id = result.get('id', None)
            if pipeline_grp_id is None:
                self.logger.error('Failed to create pipeline group.')
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.USER_ERROR,
                                                    error_desc='Failed to create Pipeline Group',
                                                    response=response)

            pipeline_grp_id = str(pipeline_grp_id)
            self.logger.info(
                'Pipeline Group {id} has been created under domain {domain_id}.'.format(id=pipeline_grp_id,
                                                                                        domain_id=domain_id))
            return GenericResponse.parse_result(status=Response.Status.SUCCESS, entity_id=pipeline_grp_id,
                                                response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(e))
            self.logger.exception('Error occurred while trying to create a new pipeline group.')
            raise PipelineError('Error occurred while trying to create a new pipeline group.')

    def list_pipeline_groups_under_domain(self, domain_id, params=None, pagination=True):
        """
        Function to list the pipeline groups under domain
        :param domain_id: Entity identifier for domain
        :type domain_id: String
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type: JSON dict
        :param pagination: Boolean value to determine whether to return entire result set or only a portion of result defined by 'limit' under params
        :type pagination: Boolean
        :return: response list
        """
        if None in {domain_id}:
            raise Exception(f"domain_id cannot be None")
        if params is None:
            params = {"limit": 20, "offset": 0}
        url_to_list_pipeline_grp = url_builder.get_pipeline_group_base_url(self.client_config,
                                                                           domain_id) + IWUtils.get_query_params_string_from_dict(
            params=params)
        pipeline_groups_list = []
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_pipeline_grp,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                initial_msg = response.get("message", "")
                while len(result) > 0:
                    pipeline_groups_list.extend(result)
                    if not pagination:
                        break
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(
                        self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                            self.client_config['bearer_token'])).content)
                    result = response.get("result", None)
                    if result is None:
                        return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                            error_code=ErrorCode.GENERIC_ERROR,
                                                            error_desc="Error in listing pipeline groups",
                                                            response=response
                                                            )

                response["result"] = pipeline_groups_list
                response["message"] = initial_msg
            return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
        except Exception as e:
            self.logger.error("Error in listing pipeline groups")
            raise PipelineError("Error in listing pipeline groups" + str(e))

    def get_pipeline_group_details(self, domain_id, pipeline_group_id):
        """
        Gets Infoworks pipeline group details for given pipeline group id
        :param pipeline_group_id: id of the pipeline group whose details are to be fetched
        :type pipeline_group_id: String
        :param domain_id: Domain id to which the pipeline  group belongs to
        :type domain_id: String
        :return: response dict
        """
        try:
            if None in {pipeline_group_id, domain_id}:
                raise Exception(f"pipeline_group_id, domain_id cannot be None")
            response = IWUtils.ejson_deserialize(self.call_api("GET", url_builder.get_pipeline_group_base_url(
                self.client_config, domain_id) + f"{pipeline_group_id}", IWUtils.get_default_header_for_v3(
                self.client_config['bearer_token'])).content)
            result = response.get('result', {})
            if result.get('id', None) is None:
                self.logger.error('Failed to find the pipeline group details')
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.USER_ERROR,
                                                    error_desc='Failed to find the pipeline group details',
                                                    response=response)

            pipeline_group_id = str(pipeline_group_id)
            self.logger.info(
                'Successfully got the pipeline group {id} details.'.format(id=pipeline_group_id))
            return GenericResponse.parse_result(status=Response.Status.SUCCESS, entity_id=pipeline_group_id,
                                                response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(e))
            self.logger.exception('Error occurred while trying to get pipeline group details.')
            raise PipelineError('Error occurred while trying to get pipeline group details.')

    def delete_pipeline_group(self, domain_id, pipeline_group_id):
        """
        Function to delete the pipeline group
        :param pipeline_group_id: id of the pipeline group that has to be deleted
        :type pipeline_group_id: String
        :param domain_id: Domain id to which the pipeline group  belongs to
        :type domain_id: String
        :return: response dict
        """
        try:
            if None in {pipeline_group_id, domain_id}:
                raise Exception(f"pipeline_group_id, domain_id cannot be None")
            response = IWUtils.ejson_deserialize(self.call_api("DELETE", url_builder.get_pipeline_group_base_url(
                self.client_config, domain_id) + f"{pipeline_group_id}", IWUtils.get_default_header_for_v3(
                self.client_config['bearer_token'])).content)
            result = response.get('result', {})
            if result.get('id', None) is None:
                self.logger.error(f'Failed to delete the pipeline group {pipeline_group_id}')
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.USER_ERROR,
                                                    error_desc=f'Failed to delete the pipeline group {pipeline_group_id}',
                                                    response=response)

            pipeline_group_id = str(pipeline_group_id)
            self.logger.info(
                'Successfully deleted the pipeline group {id}.'.format(id=pipeline_group_id))
            return GenericResponse.parse_result(status=Response.Status.SUCCESS, entity_id=pipeline_group_id,
                                                response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(e))
            self.logger.exception('Error occurred while trying to delete the pipeline group.')
            raise PipelineError('Error occurred while trying to delete the pipeline group.')

    def update_pipeline_group(self, domain_id, pipeline_group_id, pipeline_group_config):
        """
            Function to update the pipeline group
            :param pipeline_group_id: id of the pipeline group that has to be deleted
            :type pipeline_group_id: String
            :param domain_id: Domain id to which the pipeline group belongs to
            :type domain_id: String
            :param pipeline_group_config: a JSON object containing pipeline group configurations
            :type pipeline_group_config: JSON Object
        pipeline_group_config_example = {
                    "name": "test_pg1",
                    "description": "",
                    "environment_id": "a801283f2e8077120d000b49",
                    "domain_id": "576a8508b25540b4b180927f",
                    "batch_engine": "SNOWFLAKE",
                    "run_job_on_data_plane": false,
                    "pipelines": [
                        {
                            "pipeline_id": "270a56ca0ae71c79039d79e0",
                            "version": 3,
                            "execution_order": 1,
                            "run_active_version": true
                        },
                        {
                            "pipeline_id": "266c068325500cdc8ccd6d69",
                            "version": 1,
                            "execution_order": 2,
                            "run_active_version": true
                        }
                    ],
                    "custom_tags": [],
                    "query_tag": "",
                    "snowflake_warehouse": "DEMO_WH"
                }
            :return: response dict
        """
        try:
            if pipeline_group_config is None or None in {domain_id, pipeline_group_id}:
                raise Exception(
                    f"pipeline_config cannot be None or domain_id or pipeline_group_id cannot be null")
            domain_id = pipeline_group_config["domain_id"]
            response = IWUtils.ejson_deserialize(self.call_api("PUT", url_builder.get_pipeline_group_base_url(
                self.client_config, domain_id) + f"{pipeline_group_id}", IWUtils.get_default_header_for_v3(
                self.client_config['bearer_token']),
                                                               pipeline_group_config).content)
            result = response.get('result', {})
            pipeline_grp_id = result.get('id', None)
            if pipeline_grp_id is None:
                self.logger.error('Failed to update pipeline group.')
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.USER_ERROR,
                                                    error_desc='Failed to update Pipeline Group',
                                                    response=response)

            pipeline_grp_id = str(pipeline_grp_id)
            self.logger.info(
                'Pipeline Group {id} has been updated under domain {domain_id}.'.format(id=pipeline_grp_id,
                                                                                        domain_id=domain_id))
            return GenericResponse.parse_result(status=Response.Status.SUCCESS, entity_id=pipeline_grp_id,
                                                response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(e))
            self.logger.exception('Error occurred while trying to update the pipeline group.')
            raise PipelineError('Error occurred while trying to update the pipeline group.')

    def get_accessible_pipeline_groups(self, domain_id, params=None, pagination=True):
        """
            Function to get the accessible pipeline group
            :param domain_id: Domain id to which the pipeline  group  belongs to
            :type domain_id: String
            :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
            :type: JSON dict
            :param pagination: Boolean value to determine whether to return entire result set or only a portion of result defined by 'limit' under params
            :type pagination: Boolean
            :return: response dict
        """
        if domain_id is None:
            raise Exception(f"domain_id cannot be None")
        if params is None:
            params = {"limit": 20, "offset": 0}
        url_to_get_accessible_pl_grp = url_builder.get_accessible_pipeline_groups_url(self.client_config,
                                                                                      domain_id) + IWUtils.get_query_params_string_from_dict(
            params=params)
        pipeline_groups_list = []
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_get_accessible_pl_grp,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                initial_msg = response.get("message", "")
                while len(result) > 0:
                    pipeline_groups_list.extend(result)
                    if not pagination:
                        break
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(
                        self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                            self.client_config['bearer_token'])).content)
                    result = response.get("result", None)
                    if result is None:
                        return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                            error_code=ErrorCode.GENERIC_ERROR,
                                                            error_desc="Error in listing accessible pipeline groups",
                                                            response=response
                                                            )

                response["result"] = pipeline_groups_list
                response["message"] = initial_msg
            return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
        except Exception as e:
            self.logger.error("Error in listing accessible pipeline groups")
            raise PipelineError("Error in listing accessible pipeline groups" + str(e))

    def get_list_of_advanced_config_of_pipeline_groups(self, domain_id, pipeline_group_id, params=None, pagination=True):
        """
            Function to get list of Advance Config pipeline group
            :param domain_id: Domain id to which the pipeline  group belongs to
            :type domain_id: String
            param pipeline_group_id: Entity identifier of pipeline group
            :type pipeline_group_id: String
            :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
            :type: JSON dict
            :param pagination: Boolean value to determine whether to return entire result set or only a portion of result defined by 'limit' under params
            :type pagination: Boolean
            :return: response dict
        """
        if None in {pipeline_group_id, domain_id}:
            raise Exception(f"pipeline_group_id, domain_id cannot be None")
        if params is None:
            params = {"limit": 20, "offset": 0}
        url_to_get_adv_config_pl_grp = url_builder.advance_config_under_pipeline_groups_base_url(self.client_config,
                                                                                                 domain_id,
                                                                                                 pipeline_group_id) + IWUtils.get_query_params_string_from_dict(
            params=params)
        adv_config_list = []
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_get_adv_config_pl_grp,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                initial_msg = response.get("message", "")
                while len(result) > 0:
                    adv_config_list.extend(result)
                    if not pagination:
                        break
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(
                        self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                            self.client_config['bearer_token'])).content)
                    result = response.get("result", None)
                    if result is None:
                        return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                            error_code=ErrorCode.GENERIC_ERROR,
                                                            error_desc="Error in listing adv config of pipeline groups",
                                                            response=response
                                                            )

                response["result"] = adv_config_list
                response["message"] = initial_msg
            return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
        except Exception as e:
            self.logger.error("Error in listing adv config of pipeline groups")
            raise PipelineError("Error in listing adv config of pipeline groups" + str(e))

    def modify_advanced_config_for_pipeline_group(self, domain_id, pipeline_group_id, adv_config_body,
                                                  action_type="update", key=None):
        """
        Function to add/update the adv config for the pipeline group
        :param pipeline_group_id: id of the pipeline group whose details are to be fetched
        :type pipeline_group_id: String
        :param domain_id: Domain id to which the pipeline group belongs to
        :type domain_id: String
        :param action_type: values can be either create/update. default update
        :type action_type: String
        :param adv_config_body: JSON dict
        adv_config_body_example = {
            "key" : "",
            "value": "",
            "description": "",
            "is_active": True
            }
        :param key: In case of update, name of the adv config to update
        :return: response dict
        """
        if None in {pipeline_group_id, domain_id} or adv_config_body is None:
            raise Exception(f"pipeline_group_id, domain_id and adv_config_body cannot be None")
        try:
            if action_type.lower() == "create":
                request_type = "POST"
                request_url = url_builder.advance_config_under_pipeline_groups_base_url(self.client_config, domain_id,
                                                                                        pipeline_group_id)
            else:
                request_type = "PUT"
                request_url = url_builder.advance_config_under_pipeline_groups_base_url(self.client_config, domain_id,
                                                                                        pipeline_group_id) + f"{key}"
            response = IWUtils.ejson_deserialize(
                self.call_api(request_type, request_url,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                              adv_config_body).content)
            result = response.get('result', {})
            message = response.get('message', "")
            adv_config_id = result.get('id', None)
            if adv_config_id is not None:
                adv_config_id = str(adv_config_id)
                self.logger.info(
                    'Advanced Config has been created {id}.'.format(id=adv_config_id))
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, entity_id=adv_config_id,
                                                    response=response)
            elif message == "Successfully updated Advance configuration":
                self.logger.info('Advanced Config has been updated')
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
            else:
                self.logger.error(f'Failed to {action_type} advanced config for pipeline group.')
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.USER_ERROR,
                                                    error_desc=f'Failed to {action_type} advanced config for pipeline group',
                                                    response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(e))
            self.logger.exception('Error occurred while trying to add/update adv config pipeline group.')
            raise PipelineError('Error occurred while trying to add/update adv config pipeline group.')

    def get_or_delete_advance_config_details_of_pipeline_group(self, domain_id, pipeline_group_id, key,
                                                               action_type="get"):
        """
        Gets/Deletes advance configuration of pipeline group
        :param pipeline_group_id: id of the pipeline group whose details are to be fetched
        :type pipeline_group_id: String
        :param domain_id: Domain id to which the pipeline group belongs to
        :type domain_id: String
        :param key: name of the advanced configurations
        :param action_type: values can be get/delete
        :return: response dict
        """
        try:
            if None in {pipeline_group_id, domain_id, key}:
                raise Exception(f"pipeline_group_id, domain_id, key cannot be None")
            request_type = "GET" if action_type.lower() == "get" else "DELETE"
            response = IWUtils.ejson_deserialize(
                self.call_api(request_type, url_builder.advance_config_under_pipeline_groups_base_url(
                    self.client_config, domain_id, pipeline_group_id) + f"{key}", IWUtils.get_default_header_for_v3(
                    self.client_config['bearer_token'])).content)
            result = response.get('result', {})
            message = response.get("message", "")
            if action_type.lower() == "delete" and message == "Successfully removed Advance configuration":
                self.logger.info(
                    'Successfully deleted the advanced config')
                return GenericResponse.parse_result(status=Response.Status.SUCCESS,
                                                    response=response)

            if result.get('entity_id', None) is None:
                self.logger.error('Failed to find the advance config details')
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.USER_ERROR,
                                                    error_desc='Failed to find the advance config details',
                                                    response=response)

            self.logger.info(
                'Successfully got the advanced config details.')
            return GenericResponse.parse_result(status=Response.Status.SUCCESS, entity_id=result.get('entity_id', ''),
                                                response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(e))
            self.logger.exception('Error occurred while trying to get/delete adv config details.')
            raise PipelineError('Error occurred while trying to get/delete adv config details.')

    def get_pipeline_group_job_summary(self, domain_id, pipeline_group_id, job_id):
        """
        Function to get pipeline group job summary
        :param domain_id: Entity identifier for domain
        :param pipeline_group_id: Entity identifier for pipeline group
        :param job_id: Entity identifier for job
        :return: response dict
        """
        try:
            if None in {domain_id, pipeline_group_id, job_id}:
                self.logger.error("domain id or pipeline group id or job_id cannot be None")
                raise Exception("domain id or pipeline group id or job_id cannot be None")

            response = IWUtils.ejson_deserialize(self.call_api("GET", url_builder.get_pipeline_group_job_summary_url
            (self.client_config, domain_id, pipeline_group_id, job_id),
                                                               IWUtils.get_default_header_for_v3(
                                                                   self.client_config['bearer_token']), ).content)
            result = response.get('result', None)
            if result is None:
                self.logger.error(f"Failed to get the pipeline group job summary for job {job_id} ")
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.USER_ERROR,
                                                    error_desc=f"Failed to get pipeline group job summary for job {job_id} ",
                                                    response=response)
            else:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
        except Exception as e:
            raise PipelineError(f"Failed to get pipeline group job summary for job {job_id} " + str(e))



import copy
import os
import traceback
from infoworks.sdk.utils import IWUtils
from infoworks.sdk.url_builder import get_parent_entity_url, list_domains_url, configure_pipeline_url, \
    configure_workflow_url, \
    configure_source_url, get_environment_details, get_environment_storage_details, get_environment_compute_details, \
    get_environment_interactive_compute_details, get_source_configurations_url, get_pipeline_url, \
    get_data_connection, source_info, list_users_url, list_secrets_url, get_pipeline_group_base_url, list_pipelines_url, \
    create_domain_url, get_table_configuration, list_tables_under_source, create_table_group_url
from infoworks.sdk.cicd.cicd_response import CICDResponse
import json


class Utils:
    def __init__(self, serviceaccountemail):
        self.serviceaccountemail = serviceaccountemail

    def get_secret_name_from_id(self, cicd_client, secret_id):
        secret_name = None
        get_secret_details_url = list_secrets_url(cicd_client.client_config) + '?filter={"_id":"' + secret_id + '"}'
        response = cicd_client.call_api("GET", get_secret_details_url,
                                        IWUtils.get_default_header_for_v3(cicd_client.client_config['bearer_token']))
        parsed_response = IWUtils.ejson_deserialize(response.content)
        if response.status_code == 200 and len(parsed_response.get("result", [])) > 0:
            result = parsed_response.get("result", [])

            if len(result) > 0:
                secret_name = result[0]["name"]
                cicd_client.logger.info("Found secret name {} ".format(secret_name))
                return secret_name
            else:
                cicd_client.logger.info("Secret Name is {} ".format(None))
                return None

    def get_domain_id(self, cicd_client, json_obj):
        parent_entity_url = get_parent_entity_url(cicd_client.client_config)
        response = cicd_client.call_api("GET", parent_entity_url,
                                        IWUtils.get_default_header_for_v3(cicd_client.client_config['bearer_token']),
                                        data=json_obj)
        parsed_response = IWUtils.ejson_deserialize(response.content)
        if response.status_code == 200 and len(parsed_response.get("result", [])) > 0:
            result = parsed_response.get("result", [])
            if len(result) > 0:
                domain_id = result[0]["entity_id"]
                cicd_client.logger.info("Domain ID is {} ".format(domain_id))
                return domain_id
            else:
                cicd_client.logger.info("Domain ID is {} ".format(None))
                return None

    def get_env_entities_names(self, cicd_client, environment_id, environment_compute_template_id,
                               environment_storage_id):
        env_name, storage_name, compute_name = None, None, None
        if environment_id:
            response = cicd_client.call_api("GET", get_environment_details(cicd_client.client_config, environment_id),
                                            IWUtils.get_default_header_for_v3(
                                                cicd_client.client_config['bearer_token']))
            parsed_response = IWUtils.ejson_deserialize(response.content)
            if response.status_code == 200:
                result = parsed_response.get("result", None)
                if result:
                    env_name = result["name"]
        if environment_storage_id:
            response = cicd_client.call_api("GET",
                                            get_environment_storage_details(cicd_client.client_config, environment_id,
                                                                            environment_storage_id),
                                            IWUtils.get_default_header_for_v3(
                                                cicd_client.client_config['bearer_token']))
            parsed_response = IWUtils.ejson_deserialize(response.content)
            if response.status_code == 200:
                result = parsed_response.get("result", None)
                if result:
                    storage_name = result["name"]
        if environment_compute_template_id:
            response = cicd_client.call_api("GET",
                                            get_environment_compute_details(cicd_client.client_config, environment_id,
                                                                            environment_compute_template_id),
                                            IWUtils.get_default_header_for_v3(
                                                cicd_client.client_config['bearer_token']))
            parsed_response = IWUtils.ejson_deserialize(response.content)
            if response.status_code == 200:
                result = parsed_response.get("result", None)
                if result:
                    compute_name = result["name"]
                else:
                    response = cicd_client.call_api("GET",
                                                    get_environment_interactive_compute_details(
                                                        cicd_client.client_config,
                                                        environment_id,
                                                        environment_compute_template_id),
                                                    IWUtils.get_default_header_for_v3(
                                                        cicd_client.client_config['bearer_token']))
                    parsed_response = IWUtils.ejson_deserialize(response.content)
                    if response.status_code == 200:
                        result = parsed_response.get("result", None)
                        if result:
                            compute_name = result["name"]

        return env_name, storage_name, compute_name

    def get_env_details(self, cicd_client, entity_id, entity_type, domain_id=None):
        environment_id, compute_template_id, storage_id = None, None, None
        if entity_type == "pipeline":
            get_pipeline_details_url = get_pipeline_url(cicd_client.client_config, domain_id, entity_id)
            response = cicd_client.call_api("GET", get_pipeline_details_url,
                                            IWUtils.get_default_header_for_v3(
                                                cicd_client.client_config['bearer_token']))
            parsed_response = IWUtils.ejson_deserialize(response.content)
            if response.status_code == 200 and len(parsed_response.get("result", [])) > 0:
                result = parsed_response.get("result", [])
                if len(result) > 0:
                    environment_id = result.get("environment_id", None)
                    storage_id = result.get("storage_id", None)
                    compute_template_id = result.get("compute_template_id", None)
                    cicd_client.logger.info(
                        "Environment ID is {} Storage ID is {} Compute Template Id is {}".format(environment_id,
                                                                                                 storage_id,
                                                                                                 compute_template_id))
        return environment_id, compute_template_id, storage_id

    def get_dataconnection_properties(self, sub_type, properties_obj):
        final_obj = {}
        if sub_type.lower() == "snowflake":
            for key in properties_obj.keys():
                if key == "user_name":
                    final_obj["username"] = properties_obj["user_name"]
                else:
                    final_obj[key] = properties_obj[key]
        elif sub_type.lower() == "postgres":
            for key in properties_obj.keys():
                if key == "user_name":
                    final_obj["username"] = properties_obj["user_name"]
                else:
                    final_obj[key] = properties_obj[key]
        elif sub_type.lower() == "bigquery":
            for key in properties_obj.keys():
                if key == "uploadOption":
                    final_obj["upload_option"] = properties_obj["uploadOption"]
                elif key == "serverPath":
                    final_obj["server_path"] = properties_obj["serverPath"]
                else:
                    final_obj[key] = properties_obj[key]
            if "file_details" in final_obj:
                del final_obj["file_details"]

        return final_obj

    def get_sql_pipeline_config(self, cicd_client, domain_id, pipeline_id, target_file_path):
        ip = cicd_client.client_config['ip']
        port = cicd_client.client_config['port']
        protocol = cicd_client.client_config['protocol']

        get_domain_name_url = f"{protocol}://{ip}:{port}/v3/domains/{domain_id}"
        response = cicd_client.call_api("GET", get_domain_name_url,
                                        IWUtils.get_default_header_for_v3(cicd_client.client_config['bearer_token']))
        parsed_response = IWUtils.ejson_deserialize(response.content)
        domain_name = parsed_response.get("result", {}).get("name", "")

        get_active_pipeline_versions_url = f"{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/"
        response = cicd_client.call_api("GET", get_active_pipeline_versions_url, IWUtils.get_default_header_for_v3(
            cicd_client.client_config['bearer_token']))
        parsed_response = IWUtils.ejson_deserialize(response.content)
        pipeline_details = parsed_response.get("result", {})
        pipeline_name = pipeline_details.get("name", "")
        query_tag = pipeline_details.get("query_tag", "")
        custom_tags = pipeline_details.get("custom_tags", [])
        description = pipeline_details.get("description", "")
        filename = f"{domain_name}#pipeline_{pipeline_name}.json"

        active_pipeline_version = pipeline_details.get("active_version_id")

        get_pipeline_version_config_url = f"{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/versions/{active_pipeline_version}"
        response = cicd_client.call_api("GET", get_pipeline_version_config_url, IWUtils.get_default_header_for_v3(
            cicd_client.client_config[
                'bearer_token']))
        parsed_response = IWUtils.ejson_deserialize(response.content)
        pipeline_version_details = parsed_response.get("result", {})
        version_id = pipeline_version_details.get("version", 1)
        query = pipeline_version_details.get("query", "")
        pipeline_parameters = pipeline_version_details.get("pipeline_parameters", [])
        # Check for any advance configurations
        response = cicd_client.call_api("GET", get_active_pipeline_versions_url + "configurations/advance",
                                        IWUtils.get_default_header_for_v3(
                                            cicd_client.client_config[
                                                'bearer_token']))
        parsed_response = IWUtils.ejson_deserialize(response.content)
        pipeline_advance_config_details = parsed_response.get("result", [])

        environment_id, environment_compute_template_id, environment_storage_id = self.get_env_details(
            cicd_client, pipeline_id,
            "pipeline",
            domain_id)
        env_name, storage_name, compute_name = self.get_env_entities_names(
            cicd_client, environment_id,
            environment_compute_template_id,
            environment_storage_id)

        # decoded_jwt = jwt.decode(cicd_client.client_config['bearer_token'], options={"verify_signature": False})
        # user_email = json.loads(decoded_jwt.get("sub")).get("email")
        config_obj = {
            "configuration": {
                "entity": {
                    "entity_type": "pipeline",
                    "entity_id": pipeline_id,
                    "entity_name": pipeline_details.get("name"),
                    "subEntityName": f"V{version_id}",
                    "warehouse": pipeline_details.get("snowflake_warehouse")
                },
                "pipeline_configs": {
                    "type": "sql",
                    "query": query,
                    "pipeline_parameters": pipeline_parameters,
                    "batch_engine": "SNOWFLAKE",
                    "query_tag": query_tag,
                    "description": description
                },
                "pipeline_advance_configs": pipeline_advance_config_details
            },
            "environment_configurations": {"environment_name": env_name,
                                           "environment_compute_template_name": compute_name,
                                           "environment_storage_name": storage_name}
        }
        target_file_path = os.path.join(target_file_path, "pipeline", filename)
        if filename is not None and target_file_path is not None:
            cicd_client.logger.info("{} {}".format(filename, target_file_path))
            print(f"Exporting configurations file to {target_file_path}")
            with open(target_file_path, 'w') as file_ptr:
                # contents_to_write = IWUtils.ejson_serialize(config_obj)
                json.dump(config_obj, file_ptr, indent=4)
            cicd_client.logger.info("Configurations exported successfully")
            print("Configurations exported successfully")
        return filename, config_obj

    def list_pipelines(self, cicd_client, domain_id=None, params=None):
        """
        Function to list the pipelines
        :param domain_id: Entity identified for domain
        :type domain_id: String
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type: JSON dict
        :return: response list
        """

        if None in {domain_id}:
            cicd_client.logger.error("Domain ID cannot be None")
            raise Exception("Domain ID cannot be None")
        if params is None:
            params = {"limit": 20, "offset": 0}
        url_to_list_pipelines = list_pipelines_url(cicd_client.client_config, domain_id) \
                                + IWUtils.get_query_params_string_from_dict(params=params)

        pipelines_list = []
        try:
            response = IWUtils.ejson_deserialize(
                cicd_client.call_api("GET", url_to_list_pipelines,
                                     IWUtils.get_default_header_for_v3(
                                         cicd_client.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                initial_msg = response.get("message", "")
                while len(result) > 0:
                    pipelines_list.extend(result)
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=cicd_client.client_config['ip'],
                                                                      port=cicd_client.client_config['port'],
                                                                      protocol=cicd_client.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(
                        cicd_client.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                            cicd_client.client_config['bearer_token'])).content)
                    result = response.get("result", None)
                    if result is None:
                        return response
                response["result"] = pipelines_list
                response["message"] = initial_msg
            return response
        except Exception as e:
            cicd_client.logger.error("Error in listing pipelines")
            raise Exception("Error in listing pipelines" + str(e))

    def add_secret_name_to_id(self, data, cicd_client):
        for key, value in data.items():
            if isinstance(value, dict):
                # If the value is a dictionary, recurse into it
                secret_name = self.add_secret_name_to_id(value, cicd_client)
                if secret_name:
                    return secret_name
            elif key == "secret_id":
                # print(f"Found secret_name: {value}")
                # resolve secret name to ID
                secret_name = self.get_secret_name_from_id(cicd_client=cicd_client, secret_id=value)
                if secret_name:
                    data["secret_name"] = secret_name
                # Add your logic here to do something with the secret_name
                return value

    def add_custom_tags_to_id(self, data, cicd_client):
        data["custom_tag_key_values"] = {}
        for key, value in data.items():
            if key == "custom_tags":
                for tag_id in value:
                    try:
                        response = cicd_client.get_custom_tag(custom_tag_id=tag_id)
                        custom_tag_key = response['result']['response']['result']['key']
                        custom_tag_value = response['result']['response']['result']['value']
                        data["custom_tag_key_values"][custom_tag_key] = custom_tag_value
                    except Exception as error:
                        cicd_client.logger.error(f"Failed to get custom tag : {error}")
                        print(f"Failed to get custom tag : {error}")

    def dump_to_file(self, cicd_client, entity_type, domain_id, entity_id, replace_words, target_file_path,
                     dump_watermarks=True,custom_tag_id=None):
        response_to_return = {}
        filename = None
        environment_id, environment_compute_template_id, environment_storage_id = None, None, None
        if entity_type == "pipeline":
            url_to_config = configure_pipeline_url(cicd_client.client_config, domain_id, entity_id)
        elif entity_type == "workflow":
            url_to_config = configure_workflow_url(cicd_client.client_config, domain_id, entity_id)
        elif entity_type == "source":
            url_to_config = configure_source_url(cicd_client.client_config, entity_id)
        elif entity_type == "pipeline_group":
            url_to_config = get_pipeline_group_base_url(cicd_client.client_config, domain_id) + f"{entity_id}"
        else:
            return None
        cicd_client.logger.info(
            "URL to get configurations of entity {} of type {} is {}".format(entity_id, entity_type, url_to_config))
        response = cicd_client.call_api("GET", url_to_config, IWUtils.get_default_header_for_v3(
            cicd_client.client_config[
                'bearer_token']))

        parsed_response = IWUtils.ejson_deserialize(response.content)

        if response.status_code == 200:
            status = "SUCCESS"
        # removing below elif because sql pipeline migration is supported with config-migration api from 5.5
        # elif response.status_code == 406:
        #    return self.get_sql_pipeline_config(cicd_client, domain_id, entity_id, target_file_path)
        else:
            status = "FAILED"
            print(parsed_response)

        response_to_return["get_configuration_entity_response"] = CICDResponse.parse_result(status=status,
                                                                                            entity_id=entity_id,
                                                                                            response=parsed_response)
        configuration_obj = parsed_response.get('result', {})
        if len(configuration_obj) > 0:
            if entity_type == "source":
                get_src_details_url = source_info(cicd_client.client_config, entity_id)
                response = cicd_client.call_api("GET", get_src_details_url,
                                                IWUtils.get_default_header_for_v3(
                                                    cicd_client.client_config['bearer_token']))
                parsed_response = IWUtils.ejson_deserialize(response.content)
                if response.status_code == 200:
                    status = "SUCCESS"
                else:
                    status = "FAILED"
                    print("Get Source Info failed " + json.dumps(response))
                response_to_return["get_source_details_response"] = CICDResponse.parse_result(status=status,
                                                                                              entity_id=entity_id,
                                                                                              response=parsed_response)
                if response.status_code == 200 and len(parsed_response.get("result", [])) > 0:
                    result = parsed_response.get("result", [])
                    if len(result) > 0:
                        data_lake_path = result["data_lake_path"]
                        configuration_obj["configuration"]["source_configs"]["data_lake_path"] = data_lake_path
                        configuration_obj["filter_tables_properties"] = result.get("filter_tables_properties", {})
                        source_connection_objects = configuration_obj["configuration"]["source_configs"]["connection"]
                        self.add_secret_name_to_id(data=source_connection_objects, cicd_client=cicd_client)
                        # Added For Source Tags
                        source_configs = configuration_obj['configuration']['source_configs']
                        self.add_custom_tags_to_id(data=source_configs, cicd_client=cicd_client)
                        # Added for table Tags
                        table_configs = configuration_obj['configuration']['table_configs']
                        for table in table_configs:
                            self.add_custom_tags_to_id(data=table['configuration']['configuration'],
                                                       cicd_client=cicd_client)
                        # Added for table group Tags
                        table_group_configs = configuration_obj['configuration']['table_group_configs']
                        for table_group in table_group_configs:
                            self.add_custom_tags_to_id(data=table_group['configuration'], cicd_client=cicd_client)

                        # handle associated domains
                        if configuration_obj.get("configuration", {}).get("source_configs", {}).get(
                                "associated_domains", None) is None:
                            associated_domains = result.get("associated_domains", [])
                            if associated_domains:
                                configuration_obj["configuration"]["source_configs"][
                                    "associated_domains"] = associated_domains
                # Check if any table has export configurations. Works for postgres/snowflake/synapse
                # Did not test for cosmos,delimited
                for table_config in configuration_obj["configuration"]["table_configs"]:
                    if len(table_config["configuration"].get("export_configuration", [])) > 0:
                        if table_config["configuration"]["export_configuration"].get("target_type", "") in ["SNOWFLAKE",
                                                                                                            "POSTGRES",
                                                                                                            "AZURE_SQL_DW"]:
                            table_config["configuration"]["export_configuration"]["connection"]["password"] = None
                        else:
                            pass  # TO_DO

                # Dump table watermarks
                table_watermark_mappings = {}
                if dump_watermarks:
                    for table_config in configuration_obj["configuration"]["table_configs"]:
                        # table_name = table_config.get("configuration", {}).get("name", "")
                        table_id = table_config.get("entity_id", None)
                        source_id = entity_id
                        if source_id is not None and table_id is not None:
                            get_tbl_details_url = get_table_configuration(cicd_client.client_config, source_id,
                                                                          table_id)
                            response = cicd_client.call_api("GET", get_tbl_details_url,
                                                            IWUtils.get_default_header_for_v3(
                                                                cicd_client.client_config['bearer_token']))
                            parsed_response = IWUtils.ejson_deserialize(response.content)
                            if response.status_code == 200:
                                last_ingested_cdc_value = parsed_response.get("result").get("last_ingested_cdc_value",
                                                                                            None)
                                last_merged_watermark = parsed_response.get("result").get("last_merged_watermark",
                                                                                          None)
                                max_modified_timestamp = parsed_response.get("result").get("max_modified_timestamp",
                                                                                           None)
                                row_count = parsed_response.get("result").get("row_count", None)
                                full_load_performed = parsed_response.get("result").get("full_load_performed", None)
                                table_watermark_mappings[table_id] = {
                                    "last_ingested_cdc_value": last_ingested_cdc_value,
                                    "last_merged_watermark": last_merged_watermark,
                                    "row_count": row_count,
                                    "full_load_performed": full_load_performed}
                                if max_modified_timestamp:
                                    table_watermark_mappings[table_id]["max_modified_timestamp"] = max_modified_timestamp
                                configuration_obj["table_watermark_mappings"] = table_watermark_mappings
                            else:
                                print("Get Table Config Failed " + json.dumps(response))
                # filter tables and table groups by custom tag if custom tag is not None
                if custom_tag_id:
                    filtered_table_ids =[]
                    filtered_table_group_ids = []
                    list_tables_under_source_url = list_tables_under_source(config=cicd_client.client_config, source_id = entity_id) +'?filter={"configuration.custom_tags":\"'+custom_tag_id+'\"}'
                    custom_tag_tables = cicd_client.call_api("GET", list_tables_under_source_url, IWUtils.get_default_header_for_v3(
                        cicd_client.client_config[
                            'bearer_token']))

                    filtered_tables_parsed_response = IWUtils.ejson_deserialize(custom_tag_tables.content)
                    print(filtered_tables_parsed_response)
                    if custom_tag_tables.status_code == 200:
                        status = "SUCCESS"
                        filtered_tables_parsed_response=filtered_tables_parsed_response.get("result",[])
                        filtered_table_ids = [table["id"] for table in filtered_tables_parsed_response]
                        print("filtered_table_ids",filtered_table_ids)
                    else:
                        status = "FAILED"
                        print(parsed_response)

                    list_table_groups_under_source_url = create_table_group_url(config=cicd_client.client_config, source_id = entity_id) + '?filter={"configuration.custom_tags":\"'+custom_tag_id+'\"}'
                    custom_tag_table_groups = cicd_client.call_api("GET", list_table_groups_under_source_url, IWUtils.get_default_header_for_v3(
                        cicd_client.client_config[
                            'bearer_token']))
                    print(list_table_groups_under_source_url)
                    print(custom_tag_table_groups)
                    custom_tag_table_groups_parsed_response = IWUtils.ejson_deserialize(custom_tag_table_groups.content)
                    print("custom_tag_table_groups_parsed_response",custom_tag_table_groups_parsed_response)
                    if custom_tag_table_groups.status_code == 200:
                        status = "SUCCESS"
                        filtered_table_groups=custom_tag_table_groups_parsed_response.get("result",[])
                        filtered_table_group_ids = [tg["id"] for tg in filtered_table_groups]
                        print("filtered_table_group_ids",filtered_table_group_ids)
                    else:
                        status = "FAILED"
                        print(parsed_response)
                    updated_tables = []
                    updated_table_groups = []
                    for table in configuration_obj["configuration"]["table_configs"]:
                        if table["entity_id"] in filtered_table_ids:
                            updated_tables.append(table)
                        else:
                            print(f"Removing {table['configuration']['name'].upper()} from table_configs section")
                    configuration_obj["configuration"]["table_configs"] = updated_tables
                    for tg in configuration_obj["configuration"]["table_group_configs"]:
                        if tg["entity_id"] in filtered_table_group_ids:
                            updated_table_groups.append(tg)
                        else:
                            print(f"Removing {tg['configuration']['name'].upper()} from table_group_configs section")
                    configuration_obj["configuration"]["table_group_configs"] = updated_table_groups
                    updated_iw_mappings = []
                    for iw_mapping in configuration_obj["configuration"]["iw_mappings"]:
                        if iw_mapping.get("entity_type","")=="table":
                            if iw_mapping.get("entity_id", "") in filtered_table_ids:
                                updated_iw_mappings.append(iw_mapping)
                            else:
                                table_id = iw_mapping.get("entity_id", "")
                                print(
                                    f"Removing table {table_id} from iw_mappings section")
                        elif iw_mapping.get("entity_type","")=="table_group":
                            if iw_mapping.get("entity_id", "") in filtered_table_group_ids:
                                updated_iw_mappings.append(iw_mapping)
                            else:
                                tg_id = iw_mapping.get("entity_id", "")
                                print(
                                    f"Removing table group {tg_id} from iw_mappings section")
                        else:
                            updated_iw_mappings.append(iw_mapping)
                    configuration_obj["configuration"]["iw_mappings"] = updated_iw_mappings
                # add domain names to mapped domain ids
                accessible_domain_ids = configuration_obj["configuration"]["source_configs"].get("associated_domains",
                                                                                                 [])
                accessible_domain_names = []
                for domain_id in accessible_domain_ids:
                    domain_id_response = cicd_client.call_api("GET",
                                                              create_domain_url(
                                                                  cicd_client.client_config) + "/" + domain_id,
                                                              IWUtils.get_default_header_for_v3(
                                                                  cicd_client.client_config['bearer_token']))
                    domain_id_parsed_response = IWUtils.ejson_deserialize(domain_id_response.content)
                    # print(domain_id_parsed_response)
                    if domain_id_response.status_code == 200 and len(domain_id_parsed_response.get("result", [])) > 0:
                        result = domain_id_parsed_response.get("result", [])
                        if len(result) > 0:
                            domain_name = result.get("name", None)
                            if domain_name is not None:
                                accessible_domain_names.append(domain_name)
                    if len(accessible_domain_names) > 0:
                        configuration_obj["configuration"]["source_configs"][
                            "associated_domain_names"] = accessible_domain_names

                steps_to_run = {"configure_csv_source": True, "import_source_configuration": True,
                                "configure_rdbms_source_connection": True,
                                "test_source_connection": True, "browse_source_tables": True,
                                "add_tables_to_source": True,
                                "configure_tables_and_tablegroups": True}
                configuration_obj["steps_to_run"] = steps_to_run
                filename = "{}_".format(entity_type) + configuration_obj["configuration"]["entity"][
                    "entity_name"] + ".json"
                target_file_path = os.path.join(target_file_path, entity_type, filename)
                response = cicd_client.call_api("GET",
                                                get_source_configurations_url(cicd_client.client_config, entity_id),
                                                IWUtils.get_default_header_for_v3(
                                                    cicd_client.client_config['bearer_token']))
                parsed_response = IWUtils.ejson_deserialize(response.content)
                if response.status_code == 200:
                    status = "SUCCESS"
                    result = parsed_response.get("result", None)
                    if result:
                        environment_id = result["environment_id"]
                        environment_storage_id = result["storage_id"]
                else:
                    status = "FAILED"
                    print("Get Source Configurations failed " + json.dumps(response))
                response_to_return["get_source_configurations_response"] = CICDResponse.parse_result(status=status,
                                                                                                     entity_id=entity_id,
                                                                                                     response=parsed_response)
            else:
                entity_name = configuration_obj["configuration"]["entity"][
                    "entity_name"] if entity_type != "pipeline_group" else configuration_obj["name"]
                if entity_type == "pipeline":
                    # Added for Pipeline Tags
                    pipeline_configs = configuration_obj['configuration']['pipeline_configs']
                    self.add_custom_tags_to_id(data=pipeline_configs, cicd_client=cicd_client)

                    environment_id, environment_compute_template_id, environment_storage_id = self.get_env_details(
                        cicd_client, entity_id,
                        entity_type,
                        domain_id)
                    if any([environment_id, environment_compute_template_id, environment_storage_id]):
                        status = "SUCCESS"
                    else:
                        status = "FAILED"
                        print("Get Env Details failed " + json.dumps(response))
                    response = f"Found environment details: environment_id {environment_id}, environment_compute_template_id {environment_compute_template_id}, environment_storage_id {environment_storage_id}"
                    response_to_return["get_env_details_response"] = CICDResponse.parse_result(status=status,
                                                                                               entity_id=entity_id,
                                                                                               response=response)
                    # Check if there are any data connections
                    list_of_dataconnections = [item for item in configuration_obj["configuration"]["iw_mappings"] if
                                               item["entity_type"].lower() == "data_connection"]
                    if len(list_of_dataconnections) > 0:
                        configuration_obj["dataconnection_configurations"] = []
                        for dataconn in list_of_dataconnections:
                            dataconnection_obj = {}
                            entity_id = dataconn["entity_id"]
                            get_data_connection_url = get_data_connection(cicd_client.client_config,
                                                                          entity_id)
                            response = cicd_client.call_api("GET", get_data_connection_url,
                                                            IWUtils.get_default_header_for_v3(
                                                                cicd_client.client_config['bearer_token']))
                            parsed_response = IWUtils.ejson_deserialize(response.content)
                            if response.status_code == 200:
                                status = "SUCCESS"
                            else:
                                status = "FAILED"
                                print("Get Data Connection Details failed " + json.dumps(response))
                                cicd_client.logger.error("Get Data Connection Details failed " + json.dumps(response))

                            response_to_return["get_data_connection_details_response"] = CICDResponse.parse_result(
                                status=status,
                                entity_id=entity_id,
                                response=parsed_response)
                            if response.status_code == 200 and len(parsed_response.get("result", [])) > 0:
                                result = parsed_response.get("result", [])
                                if len(result) > 0:
                                    result["properties"] = self.get_dataconnection_properties(result["sub_type"],
                                                                                              result["properties"])
                                for key in ['name', 'type', 'sub_type', 'properties']:
                                    dataconnection_obj[key] = result[key]
                                configuration_obj["dataconnection_configurations"].append(
                                    copy.deepcopy(dataconnection_obj))
                elif entity_type == "pipeline_group":
                    pipelines_under_domain_parsed_response = self.list_pipelines(cicd_client, domain_id=domain_id)
                    pipeline_name_lookup = {}
                    if pipelines_under_domain_parsed_response.get("result", "") == "":
                        cicd_client.logger.error(f"Failed to list the pipelines under domain {domain_id}")
                        cicd_client.logger.error(pipelines_under_domain_parsed_response)
                        print(f"Failed to list the pipelines under domain {domain_id}")
                        print(pipelines_under_domain_parsed_response)
                        raise Exception(f"Failed to list the pipelines under domain {domain_id}")
                    for pipeline in pipelines_under_domain_parsed_response["result"]:
                        pipeline_name_lookup[pipeline["id"]] = pipeline["name"]
                    for index, pipeline in enumerate(configuration_obj["pipelines"]):
                        pipeline["name"] = pipeline_name_lookup.get(pipeline["pipeline_id"], None)
                    environment_id = configuration_obj["environment_id"]
                    # Added for Pipeline Group Tags
                    pipeline_group_configs = configuration_obj
                    self.add_custom_tags_to_id(data=pipeline_group_configs, cicd_client=cicd_client)

                domains_url_base = list_domains_url(cicd_client.client_config)
                filter_condition = IWUtils.ejson_serialize({"_id": domain_id})
                domains_url = domains_url_base + f"?filter={{filter_condition}}".format(
                    filter_condition=filter_condition)
                response = cicd_client.call_api("GET", domains_url,
                                                IWUtils.get_default_header_for_v3(
                                                    cicd_client.client_config['bearer_token']))
                parsed_response = IWUtils.ejson_deserialize(response.content)
                if response.status_code == 200:
                    status = "SUCCESS"
                    cicd_client.logger.info("Got domain details successfully")
                else:
                    status = "FAILED"
                    print("Get Domains Details failed " + json.dumps(response))
                response_to_return["get_domain_details_response"] = CICDResponse.parse_result(status=status,
                                                                                              entity_id=entity_id,
                                                                                              response=parsed_response)
                existing_domain_name = None
                if response.status_code == 200 and len(parsed_response.get("result", [])) > 0:
                    result = parsed_response.get("result", [])
                    existing_domain_name = result[0]["name"]
                    if entity_type == "workflow":
                        environment_ids = result[0]["environment_ids"]
                if existing_domain_name:
                    filename = existing_domain_name + "#{}_".format(entity_type) + \
                               entity_name + ".json"
                    target_file_path = os.path.join(target_file_path, entity_type, filename)

            if entity_type == "workflow":
                storage_name, compute_name = None, None
                environment_names = []
                if len(environment_ids) > 0:
                    for environment_id in environment_ids:
                        env_name, storage_name, compute_name = self.get_env_entities_names(cicd_client, environment_id,
                                                                                           None, None)
                        environment_names.append(env_name)
                        print(env_name, storage_name, compute_name)
                    configuration_obj["environment_configurations"] = {"environment_name": environment_names,
                                                                       "environment_compute_template_name": compute_name,
                                                                       "environment_storage_name": storage_name}
                else:
                    print("in else")
                # Added for Workflow Tags
                workflow_configs = configuration_obj['configuration']['workflow']
                self.add_custom_tags_to_id(data=workflow_configs, cicd_client=cicd_client)
            else:
                env_name, storage_name, compute_name = self.get_env_entities_names(
                    cicd_client, environment_id,
                    environment_compute_template_id,
                    environment_storage_id)

                configuration_obj["environment_configurations"] = {"environment_name": env_name,
                                                                   "environment_compute_template_name": compute_name,
                                                                   "environment_storage_name": storage_name}

            if entity_type != "pipeline_group":
                filter_condition = IWUtils.ejson_serialize(
                    {"$or": [{"_id": configuration_obj["configuration"]["export"]["exported_by"]},
                             {"profile.name": configuration_obj["configuration"]["export"]["exported_by"]}]})
                url_to_list_users_base = list_users_url(cicd_client.client_config)
                url_to_list_users = url_to_list_users_base + f"?filter={{filter_condition}}".format(
                    filter_condition=filter_condition)
                response = cicd_client.call_api("GET", url_to_list_users,
                                                IWUtils.get_default_header_for_v3(
                                                    cicd_client.client_config['bearer_token']))
                parsed_response = IWUtils.ejson_deserialize(response.content)
                configuration_obj["user_email"] = self.serviceaccountemail
                if response.status_code == 200 and len(parsed_response.get("result", [])) > 0:
                    result = parsed_response.get("result", [])
                    configuration_obj["user_email"] = result[0]["profile"].get("email", "admin@infoworks.io")
            try:
                if filename is not None and target_file_path is not None:
                    cicd_client.logger.info("{} {}".format(filename, target_file_path))
                    print(f"Exporting configurations file to {target_file_path}")
                    with open(target_file_path, 'w') as file_ptr:
                        # contents_to_write = IWUtils.ejson_serialize(configuration_obj)
                        # if replace_words != "":
                        #    for key, value in [item.split("->") for item in replace_words.split(";")]:
                        #        contents_to_write = contents_to_write.replace(key, value)
                        # file_ptr.write(contents_to_write)
                        json.dump(configuration_obj, file_ptr, indent=4)
                    cicd_client.logger.info("Configurations exported successfully")
                    print("Configurations exported successfully")
            except Exception as e:
                cicd_client.logger.error(str(e))
                print(str(e))
        else:
            cicd_client.logger.info("Unable to export the configurations")
            print("Unable to export the configurations")
        # for item in response_to_return:
        #    print(item, json.dumps(response_to_return[item]))
        return filename, configuration_obj


from infoworks.error import JobsError
from infoworks.sdk import url_builder
from infoworks.sdk.base_client import BaseClient
from infoworks.sdk.generic_response import GenericResponse
from infoworks.sdk.local_configurations import Response, ErrorCode
from infoworks.sdk.utils import IWUtils


class JobsClient(BaseClient):

    def __init__(self):
        super(JobsClient, self).__init__()

    def get_job_details(self, job_id=None, params=None, pagination=True):
        """
        Function to get the job details
        :param job_id: entity identifier for job
        :type: String
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type: JSON dict
        :param pagination: Boolean value to determine whether to return entire result set or only a portion of result defined by 'limit' under params
        :type pagination: Boolean
        :return: response list of dict
        """
        if params is None:
            params = {"limit": 20, "offset": 0}

        url_to_list_jobs = url_builder.get_jobs_url(self.client_config)
        if job_id is not None:
            url_to_list_jobs = url_to_list_jobs + f"/{job_id}"
        else:
            url_to_list_jobs = url_to_list_jobs + IWUtils.get_query_params_string_from_dict(params=params)
        job_details = []

        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_jobs,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            initial_msg = ""
            if response is not None:
                initial_msg = response.get("message", "")
                result = response.get("result", None)
                if result is None:
                    self.logger.error('Failed to get job details')
                    return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                        error_desc='Failed to get job details',
                                                        response=response)
                if job_id is not None:
                    job_details.extend([result])
                else:
                    while len(result) > 0:
                        job_details.extend(result)
                        if not pagination:
                            break
                        nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                          ip=self.client_config['ip'],
                                                                          port=self.client_config['port'],
                                                                          protocol=self.client_config['protocol'],
                                                                          )
                        response = IWUtils.ejson_deserialize(
                            self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                                self.client_config['bearer_token'])).content)
                        result = response.get("result", [])
            response["result"] = job_details
            response["message"] = initial_msg
            return GenericResponse.parse_result(job_id=job_id, status=Response.Status.SUCCESS, response=response)
        except Exception as e:
            self.logger.error("Error in getting job details")
            raise JobsError("Error in getting job details" + str(e))

    def resubmit_failed_tables_for_ingestion(self, job_id=None, role="admin"):
        """
        Resubmit the failed tables for Ingestion
        :param job_id: job id to resubmit the failed tables for ingestion
        :type job_id: String
        :param role: Can be either admin/prodops
        :return: response dict
        """
        if None in {job_id}:
            self.logger.error("job_id cannot be None")
            raise Exception("job_id cannot be None")
        url_to_resubmit_job_for_failed_tables = url_builder.get_jobs_url(self.client_config) + f"/{job_id}/resubmit"
        if role.lower() == "prodops":
            url_to_resubmit_job_for_failed_tables.reaplce("admin", "prodops")
        response = None
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("POST", url_to_resubmit_job_for_failed_tables,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)

            result = response.get('result', [])
            if len(result) == 0:
                self.logger.error('Failed to Resubmit the failed tables for Ingestion')
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_desc='Failed to Resubmit the failed tables for Ingestion',
                                                    response=response)
            new_job_id = result.get('id', None)
            if result.get('id', None) is None:
                self.logger.error(f'Failed to resubmit the job {job_id}')
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_desc=f'Failed to resubmit the job {job_id}',
                                                    response=response)

            self.logger.info(
                'Successfully resubmitted the job {job_id} with new job id {new_job_id}.'.format(job_id=job_id,
                                                                                                 new_job_id=new_job_id))
            return GenericResponse.parse_result(status=Response.Status.SUCCESS, job_id=new_job_id,
                                                response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to trigger workflow.')
            raise JobsError('Error occurred while trying to trigger workflow.')

    def get_job_logs_as_text_stream(self, job_id=None, params=None):
        """
        Get the job logs as text stream
        :param job_id: job id to get logs
        :type job_id: String
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type: JSON dict
        :return: response dict
        """
        if None in {job_id}:
            self.logger.error("job_id cannot be None")
            raise Exception("job_id cannot be None")
        if params is None:
            params = {"limit": 20, "offset": 0}
        url_to_get_job_logs_text = url_builder.get_jobs_url(self.client_config) + f"/{job_id}/logs"
        url_to_get_job_logs_text = url_to_get_job_logs_text + IWUtils.get_query_params_string_from_dict(params=params)
        response = None
        try:
            headers = IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
            headers["Content-Type"] = "application/octet-stream"

            response = self.call_api("GET", url_to_get_job_logs_text,
                                     headers)
            result = response.content
            if response.status_code != 200:
                self.logger.error(f'Unable to get logs for job {job_id}')
                return {"status": Response.Status.FAILED,
                        "error_desc": f'Unable to get logs for the job {job_id}',
                        "response": result}

            self.logger.info(
                'Successfully got the logs for the job {job_id}.'.format(job_id=job_id))
            return {"status": Response.Status.SUCCESS, "job_id": job_id,
                    "response": result}

        except Exception as e:
            self.logger.error('Response from server: ' + str(response.content))
            self.logger.exception('Error occurred while trying to get logs for job.')
            raise JobsError('Error occurred while trying to get logs for job.')

    def get_cluster_job_details(self, job_id=None, run_id=None, params=None, pagination=True):
        """
        Function to get cluster job logs for iw_job using job_id
        :param job_id: job_id for the job
        :type job_id: String
        :param run_id: run_id for the job
        :type run_id: String
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type: JSON dict
        :param pagination: Boolean value to determine whether to return entire result set or only a portion of result defined by 'limit' under params
        :type pagination: Boolean
        :return: response dict
        """
        if None in {job_id}:
            self.logger.error("job_id cannot be None")
            raise Exception("job_id cannot be None")
        if params is None:
            params = {"limit": 20, "offset": 0}

        url_to_get_cluster_job_details = url_builder.get_jobs_url(self.client_config) + f"/{job_id}/runs"
        if run_id:
            url_to_get_cluster_job_details = url_to_get_cluster_job_details + f"/{run_id}"
        else:
            url_to_get_cluster_job_details = url_to_get_cluster_job_details + IWUtils.get_query_params_string_from_dict(
                params=params)
        job_details = []
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_get_cluster_job_details,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            initial_msg = ""
            if response is not None:
                initial_msg = response.get("message", "")
                result = response.get("result", None)
                if result is None:
                    self.logger.error('Failed to get job details')
                    return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                        error_desc='Failed to get cluster job details',
                                                        response=response)
                if run_id is not None:
                    job_details.extend([result])
                else:
                    while len(result) > 0:
                        job_details.extend(result)
                        if not pagination:
                            break
                        nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                          ip=self.client_config['ip'],
                                                                          port=self.client_config['port'],
                                                                          protocol=self.client_config['protocol'],
                                                                          )
                        response = IWUtils.ejson_deserialize(
                            self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                                self.client_config['bearer_token'])).content)
                        result = response.get("result", [])
            response["result"] = job_details
            response["message"] = initial_msg
            return GenericResponse.parse_result(job_id=job_id, status=Response.Status.SUCCESS, response=response)
        except Exception as e:
            self.logger.error("Error in getting cluster job details")
            raise JobsError("Error in getting cluster job details" + str(e))

    def get_admin_job_details(self, params=None, pagination=True):
        """
        Function to get the admin job details
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type: JSON dict
        :param pagination: Boolean value to determine whether to return entire result set or only a portion of result defined by 'limit' under params
        :type pagination: Boolean
        :return: response list of dict
        """
        if params is None:
            params = {"limit": 20, "offset": 0}

        url_to_list_jobs = url_builder.get_admin_jobs_url(self.client_config)
        url_to_list_jobs = url_to_list_jobs + IWUtils.get_query_params_string_from_dict(params=params)
        job_details = []

        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_jobs,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            initial_msg = ""
            if response is not None:
                initial_msg = response.get("message", "")
                result = response.get("result", [])
                if len(result) == 0:
                    self.logger.error(f"Failed to get the admin job details.")
                    return GenericResponse.parse_result(status=Response.Status.FAILED, error_code=ErrorCode.USER_ERROR,
                                                        error_desc=f"Failed to get the admin job details.",
                                                        response=response)
                while len(result) > 0:
                    job_details.extend(result)
                    if not pagination:
                        break
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(
                        self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                            self.client_config['bearer_token'])).content)
                    result = response.get("result", [])
            response["result"] = job_details
            response["message"] = initial_msg
            return GenericResponse.parse_result(job_id=None, status=Response.Status.SUCCESS, response=response)
        except Exception as e:
            self.logger.error("Error in getting job details")
            raise JobsError("Error in getting job details" + str(e))

    def get_all_jobs_for_source(self, source_id=None, params=None, pagination=True):
        """
        Function to get all jobs for a particular source
        :param source_id: entity identifier for which the jobs are to be fetched
        :type: String
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type: JSON dict
        :param pagination: Boolean value to determine whether to return entire result set or only a portion of result defined by limit under params
        :type pagination: Boolean
        :return: response list of dict
        """
        if None in {source_id}:
            self.logger.error("source_id cannot be None")
            raise Exception("source_id cannot be None")
        if params is None:
            params = {"limit": 20, "offset": 0}

        url_to_list_jobs = url_builder.get_source_details_url(self.client_config)
        if source_id is not None:
            url_to_list_jobs = url_to_list_jobs + f"/{source_id}/jobs"
        else:
            self.logger.error("Pass the mandatory parameter source_id for this method")
            raise JobsError("Pass the mandatory parameter source_id for this method")
        url_to_list_jobs = url_to_list_jobs + IWUtils.get_query_params_string_from_dict(params=params)
        job_details = []
        initial_msg = ""
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_jobs,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                initial_msg = response.get("message", "")
                result = response.get("result", None)
                if result is None:
                    self.logger.error(f"Failed to get the source jobs details.")
                    return GenericResponse.parse_result(status=Response.Status.FAILED, error_code=ErrorCode.USER_ERROR,
                                                        error_desc=f"Failed to get the source jobs details.",
                                                        response=response)

                while len(result) > 0:
                    job_details.extend(result)
                    if not pagination:
                        break
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(
                        self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                            self.client_config['bearer_token'])).content)
                    result = response.get("result", [])
            response["message"] = initial_msg
            response["result"] = job_details
            return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
        except Exception as e:
            self.logger.error("Error in getting job details")
            raise JobsError("Error in getting job details" + str(e))

    def get_source_job_summary_or_logs(self, job_id=None, source_id=None, type="summary", num_of_lines=1000):
        """
        Function to get job summary for given job_id
        :param source_id: source_id for the job
        :type source_id: String
        :param job_id: job_id for the job
        :type job_id: String
        :param type: can be either summary/logs
        :param num_of_lines: Number of log lines to capture
        :return: response dict
        """
        if None in {job_id, source_id}:
            self.logger.error("job_id or source_id cannot be None")
            raise Exception("job_id or source_id cannot be None")
        try:
            if type.lower() == "summary":
                url_to_get_cluster_job_details = url_builder.get_source_details_url(
                    self.client_config) + f"/{source_id}/jobs/{job_id}/summary"
            else:
                url_to_get_cluster_job_details = url_builder.get_source_details_url(
                    self.client_config) + f"/{source_id}/jobs/{job_id}/logs?num_of_lines={num_of_lines}"
            response = IWUtils.ejson_deserialize(self.call_api("GET", url_to_get_cluster_job_details,
                                                               IWUtils.get_default_header_for_v3(
                                                                   self.client_config['bearer_token']),
                                                               ).content)
            result = response.get('result', None)
            if result is None:
                self.logger.error(f"Failed to get the crawl job summary/log for job_id {job_id}.")
                return GenericResponse.parse_result(status=Response.Status.FAILED, error_code=ErrorCode.USER_ERROR,
                                                    error_desc=f"Failed to get the crawl job summary/log for job_id {job_id}.",
                                                    response=response, job_id=job_id)
            else:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
        except Exception as e:
            raise JobsError(f"Failed to get the crawl job summary/log for job_id {job_id}." + str(e))

    def get_interactive_jobs_list(self, source_id=None, job_id=None, params=None, pagination=True):
        """
        Function to get all interactive jobs
        :param source_id: source_id for the interactive jobs
        :type source_id: String
        :param job_id: Entity identifier of the job
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type: JSON dict
        :param pagination: Boolean value to determine whether to return entire result set or only a portion of result defined by limit under params
        :type pagination: Boolean
        :return: response list of dict
        """
        if None in {source_id}:
            self.logger.error("source_id cannot be None")
            raise Exception("source_id cannot be None")
        if params is None:
            params = {"limit": 20, "offset": 0}

        url_to_list_jobs = url_builder.get_interactive_jobs_url(self.client_config, source_id)
        if job_id:
            url_to_list_jobs = url_to_list_jobs + f"/{job_id}"
        url_to_list_jobs = url_to_list_jobs + IWUtils.get_query_params_string_from_dict(params=params)
        job_details = []

        try:
            initial_msg = ""
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_jobs,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                initial_msg = response.get("message", "")
                result = response.get("result", None)
                if result is None:
                    self.logger.error(f"Failed to get the interactive jobs list.")
                    return GenericResponse.parse_result(status=Response.Status.FAILED, error_code=ErrorCode.USER_ERROR,
                                                        error_desc=f"Failed to get the interactive jobs list.",
                                                        response=response)
                if job_id is not None:
                    job_details.extend([result])
                else:
                    while len(result) > 0:
                        job_details.extend(result)
                        if not pagination:
                            break
                        nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                          ip=self.client_config['ip'],
                                                                          port=self.client_config['port'],
                                                                          protocol=self.client_config['protocol'],
                                                                          )
                        response = IWUtils.ejson_deserialize(
                            self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                                self.client_config['bearer_token'])).content)
                        result = response.get("result", [])
            response["result"] = job_details
            response["message"] = initial_msg
            return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
        except Exception as e:
            self.logger.error("Error in getting job details")
            raise JobsError("Error in getting job details" + str(e))

    def get_list_of_pipeline_jobs(self, domain_id=None, pipeline_id=None, params=None, pagination=True):
        """
        Function to get all jobs for a particular pipeline
        :param domain_id: entity identifier for domain
        :type domain_id: String
        :param pipeline_id: entity identifier for pipeline
        :type pipeline_id: String
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type: JSON dict
        :param pagination: Boolean value to determine whether to return entire result set or only a portion of result defined by limit under params
        :type pagination: Boolean
        :return: response list of dict
        """
        if None in {domain_id, pipeline_id}:
            self.logger.error("domain_id or pipeline_id cannot be None")
            raise Exception("domain_id or pipeline_id cannot be None")
        if params is None:
            params = {"limit": 20, "offset": 0}

        url_to_list_jobs = url_builder.get_pipeline_jobs_url(self.client_config, domain_id=domain_id,
                                                             pipeline_id=pipeline_id)
        url_to_list_jobs = url_to_list_jobs + IWUtils.get_query_params_string_from_dict(params=params)
        job_details = []
        initial_msg = ""
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_jobs,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                initial_msg = response.get("message", "")
                result = response.get("result", [])
                while len(result) > 0:
                    job_details.extend(result)
                    if not pagination:
                        break
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(
                        self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                            self.client_config['bearer_token'])).content)
                    result = response.get("result", [])
            response["result"] = job_details
            response["message"] = initial_msg
            return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
        except Exception as e:
            self.logger.error("Error in getting job details")
            raise JobsError("Error in getting job details" + str(e))

    def submit_pipeline_job(self, domain_id=None, pipeline_id=None, version_id=None, job_type=None,
                            updated_pipeline_parameters=[]):
        """
        Function to initiate a pipeline job for given pipeline id
        :param domain_id: domain_id for the pipeline
        :type domain_id: String
        :param pipeline_id: Id of the pipeline
        :type pipeline_id: String
        :param version_id: version_id of the pipeline
        :type version_id: String
        :param updated_pipeline_parameters: list of pipeline parameters(Key Value pairs)
        :type updated_pipeline_parameters: Array
        :param job_type: type of job to run on the given pipeline(pipeline_build,pipeline_metadata)
        :type job_type: String
        :return: response dict
        """
        if None in {domain_id, pipeline_id}:
            self.logger.error("domain_id or pipeline_id cannot be None")
            raise Exception("domain_id or pipeline_id cannot be None")
        try:
            url_to_initiate_pipeline_job = url_builder.get_pipeline_jobs_url(self.client_config, domain_id=domain_id,
                                                                             pipeline_id=pipeline_id)
            api_payload = {}
            if version_id:
                api_payload["version_id"] = version_id
            if job_type:
                api_payload["job_type"] = job_type
            if updated_pipeline_parameters:
                api_payload["updated_pipeline_parameters"] = updated_pipeline_parameters
            response = IWUtils.ejson_deserialize(self.call_api("POST", url_to_initiate_pipeline_job,
                                                               IWUtils.get_default_header_for_v3(
                                                                   self.client_config['bearer_token']), data=api_payload
                                                               ).content)
            result = response.get('result', None)
            if result is None:
                self.logger.error(f"Failed to initiate {job_type} job for pipeline {pipeline_id}.")
                return GenericResponse.parse_result(status=Response.Status.FAILED, error_code=ErrorCode.USER_ERROR,
                                                    error_desc=f"Failed to initiate {job_type} job for pipeline {pipeline_id}.",
                                                    response=response)
            else:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
        except Exception as e:
            raise JobsError(f"Failed to initiate {job_type} job for pipeline {pipeline_id}." + str(e))

    def cancel_job(self, job_id=None, role="admin"):
        """
        Function to cancel an Infoworks Job by prodops user
        :param job_id: job_id in Infoworks
        :type job_id: String
        :param role: can be either admin/prodops
        :return: response dict
        """
        if None in {job_id}:
            self.logger.error("job_id cannot be None")
            raise Exception("job_id cannot be None")
        try:
            url_to_cancel_job = url_builder.get_cancel_job_url(self.client_config, job_id=job_id)
            if role.lower() == "admin":
                url_to_cancel_job.replace("prodops", "admin")
            response = IWUtils.ejson_deserialize(self.call_api("GET", url_to_cancel_job,
                                                               IWUtils.get_default_header_for_v3(
                                                                   self.client_config['bearer_token'])
                                                               ).content)
            result = response.get("message", "")
            if result != 'Requested Job Cancellation':
                self.logger.error(f"Failed to cancel job {job_id}.")
                return GenericResponse.parse_result(status=Response.Status.FAILED, error_code=ErrorCode.USER_ERROR,
                                                    error_desc=f"Failed to cancel job {job_id}.", response=response)
            return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
        except Exception as e:
            raise JobsError(f"Failed to cancel job for job_id {job_id}." + str(e))

    def get_job_details_by_prodops_user(self, job_id=None, params=None, pagination=True):
        """
        Function to get the job details by prodops user
        :param job_id: entity identifier for job
        :type: String
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type: JSON dict
        :param pagination: Boolean value to determine whether to return entire result set or only a portion of result defined by limit under params
        :type pagination: Boolean
        :return: response list of dict
        """
        if None in {job_id}:
            self.logger.error("job_id cannot be None")
            raise Exception("job_id cannot be None")
        if params is None:
            params = {"limit": 20, "offset": 0}

        url_to_list_jobs = url_builder.get_jobs_prodops_url(self.client_config)
        if job_id is not None:
            url_to_list_jobs = url_to_list_jobs + f"/{job_id}"
        else:
            url_to_list_jobs = url_to_list_jobs + IWUtils.get_query_params_string_from_dict(params=params)
        job_details = []

        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_jobs,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            initial_msg = ""
            if response is not None:
                initial_msg = response.get("message", "")
                result = response.get("result", None)
                if result is None:
                    self.logger.error('Failed to get job details')
                    return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                        error_desc='Failed to get job details',
                                                        response=response)
                if job_id is not None:
                    job_details.extend([result])
                else:
                    while len(result) > 0:
                        job_details.extend(result)
                        if not pagination:
                            break
                        nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                          ip=self.client_config['ip'],
                                                                          port=self.client_config['port'],
                                                                          protocol=self.client_config['protocol'],
                                                                          )
                        response = IWUtils.ejson_deserialize(
                            self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                                self.client_config['bearer_token'])).content)
                        result = response.get("result", [])
            response["result"] = job_details
            response["message"] = initial_msg
            return GenericResponse.parse_result(job_id=job_id, status=Response.Status.SUCCESS, response=response)
        except Exception as e:
            self.logger.error("Error in getting job details")
            raise JobsError("Error in getting job details" + str(e))
